REBOL [
	author: 'chl
	dates: [
		2002-12-22 "started"
		2002-12-25 "resize, workspace-drag"
		2002-12-26 "eye candy - smooth move on resize, note border, text origin"
		2002-12-28 "arrow drawing integration"
		2002-12-30 "patch-coord, refactoring, connection selection"
		2002-12-31 "refactoring, add nodes via ui, <DEL>ete nodes and connections, png snapshot (ctrl-i)"
		2003-01-01 "fixed a couple of buglets"
		2003-01-02 "restructuring & refactoring (libs), basic 3space integration - save & load"
		2003-01-06 "integrated node editor"
		2006-09-16 "fixed a variety if bitrot-induced bugs as reported on reboltalk (+ some concerning weird rebol path! behaviour)"
	]
]

; -- preferences --

prefs: construct load %3flex.preferences

; -- defaultwerk --

default-font-small: make face/font [
	name: "Verdana"
	size: 11
]

; -- libraries --

do %lib/misc-lib.r
do %lib/3space-in-memory-lib.r
do %lib/node-editor-lib.r

; -- initialization --

node-editor/init space

update-changed-node: func [id] [
	foreach face workspace/pane [
		if all [
			face/type = 'node
			face/id = id
		] [
			show face
		]
	]
]

node-editor/add-change-listener :update-changed-node

; -- edge layout functions --

side-center-coords: func [n] [ ; [l r t b]
	reduce [
		n/offset + to-pair reduce [0 to-integer n/size/y / 2]
		n/offset + to-pair reduce [n/size/x to-integer n/size/y / 2]
		n/offset + to-pair reduce [to-integer n/size/x / 2 0]
		n/offset + to-pair reduce [to-integer n/size/x / 2 n/size/y]
	]
]

patch-coord: func [c1 c2 i n /local d1 d2 r] [
	either i < 3 [
		d1: 'y 
		d2: 'x
	] [
		d1: 'x
		d2: 'y
	]
	r: reduce copy [
		min (n/offset/:d1 + n/size/:d1 - 10) 
			max (n/offset/:d1 + 10) 
				to-integer (((5 * c1/:d1) + c2/:d1) / 6)
		c1/:d2
	]
	if i < 3 [
		reverse r
	]
	to-pair r
]

optimal-anchor-coords: func [n1 n2 /local s1 s2 distances i j s i1 i2 coords] [
	distances: copy []
	s1: side-center-coords n1
	s2: side-center-coords n2
	for i 1 4 1 [
		for j 1 4 1 [
			append distances vector-length s1/:i - s2/:j
		]
	]
	i: index? find distances (min~ distances)
	coords: reduce [
		pick s1 i1: (1 + to-integer (i - 1 / 4))
		pick s2 i2: (s: i // 4 either zero? s [4] [s])
	]
	reduce [
		patch-coord coords/1 coords/2 i1 n1
		patch-coord coords/2 coords/1 i2 n2
	]
]

; -- global statewerk --

current-file: none

active-node: none
active-edge: copy []

connecting: false

edge-subject: none
edge-object: none

edges: copy []
edge-coords: copy []

; -- faces --

workspace: center-face make face [
	type: 'workspace
	color: prefs/workspace-color
	edge: 0x0
	size: initial-workspace-size
	pane: copy []
	pane-attic: copy []
	effect: copy []
	original-size: self/size
	delta-to-move: 0x0
	last-click: 0
	rate: 25
	repaint: func [/local f coords effect] [
		f: self
		effect: copy []
		clear edge-coords
		foreach edge edges [
			append edge-coords join 
				coords: optimal-anchor-coords edge/1 edge/2
				edge
			append effect draw-arrow 
				coords/1 
				coords/2 
				(
					either all reduce [
						(edge/1 = active-edge/1)
						(edge/2 = active-edge/2)
					] [
						red
					] [
						luma prefs/inactive-node-color 0.75
					]
				)
				12
				5
				f
		]
		f/effect: compose/deep [draw [(effect)]]
	]
	hide-arrows: does [
		if not none? workspace/effect [
			workspace/effect: none
			show workspace
		]
	]
]

node!: make face [
	type: 'node
	active: false
	id: none
	color: prefs/inactive-node-color
	edge: 0x0
	size: prefs/minimal-node-size
	original-size: 0x0
	dragged: false
	last-click: 0
	text: "Funky Node"
	font: make object! [
		name: "Verdana"
		style: none
		size: 11
		color: black
	]
	para: make para [
		origin: 6x4
	]
	feel: make feel [
		engage: func [f a e /local new-size mode new-pane] [
			if a = 'down [
				if 300 > (e/time - f/last-click) [ ; double-click
				 	; view/new/title layout [area f/text] "Edit Node"
					sync-model
					node-editor/switch-subject f/id
				]
				f/last-click: e/time
				activate-node f
				f/data: e/offset
				f/original-size: f/size
				if all [
					f/data/x < 10
					f/data/y > (f/size/y - 10)
				] [
					connecting: true
					edge-subject: f
					f/repaint
					show f
					workspace/image: to-image workspace
					workspace/pane-attic: copy workspace/pane
					clear workspace/pane
					workspace/hide-arrows
				]
			]
			if a = 'up [
				f/dragged: false
				if connecting [
					if not none? edge-object [
						add-edge edge-subject edge-object
					]
					connecting: false
					edge-subject/repaint
					edge-subject: none
					edge-object: none
					workspace/image: none
					workspace/repaint
					append workspace/pane workspace/pane-attic
				]
				workspace/repaint
				show workspace

			]
			if find [over away] a [
				mode: 'move
				if f/dragged or all [
					f/data/x > (f/size/x - 20)
					f/data/y > (f/size/y - 20)
				] [
					mode: 'resize
				]
				if connecting [
					mode: 'connect
				]
				switch mode [
					resize [
						f/dragged: true
						new-size: f/original-size + e/offset - f/data
						new-size/x: max new-size/x prefs/minimal-node-size/x
						new-size/y: max new-size/y prefs/minimal-node-size/y
						f/size: new-size
						workspace/hide-arrows
						f/repaint
					]
					move [
						f/offset: f/offset + e/offset - f/data
						if not (e/offset - f/data = 0x0) [
							workspace/hide-arrows
						]
					]
					connect [
						edge-object: none
						foreach face workspace/pane-attic [
							if within? (f/offset + e/offset) face/offset face/size [
								edge-object: face
							]
						]
						workspace/effect: compose/deep [
							draw [
								(draw-arrow f/offset + (f/size / 2) f/offset + e/offset red 12 5 workspace)
								(
									either none? edge-object [
										[]
									] [	
										compose [
											pen red
											fill-pen none
											box 
												(edge-object/offset) 
												(edge-object/offset + edge-object/size - 1)
										]
									]
								)
							]
						]
						show workspace
					]
				]
			]
			show f
		]
	]
	effect: copy []
	repaint: func [/local f] [
		f: self
		f/color: either f/active [prefs/active-node-color] [prefs/inactive-node-color]
		f/effect: compose/deep [
			draw [
				pen (luma (either f/active [prefs/active-node-color] [prefs/inactive-node-color]) 0.75)
				line
					(to-pair reduce [2 f/size/y - 2])
					(to-pair reduce [f/size/x - 2 f/size/y - 2])
					(to-pair reduce [f/size/x - 2 2])
				pen white
				line
					(to-pair reduce [2 f/size/y - 3])
					2x2
					(to-pair reduce [f/size/x - 2 2])
					
					
				pen red
				fill-pen red
				box
					(to-pair reduce [f/size/x - 5 f/size/y - 5])
					(to-pair reduce [f/size/x f/size/y])
				pen (luma (either f/active [prefs/active-node-color] [prefs/inactive-node-color]) 0.9)
				line
					(to-pair reduce [4 f/size/y - 12])
					(to-pair reduce [12 f/size/y - 4])

				pen white
				line
					(to-pair reduce [5 f/size/y - 13])
					(to-pair reduce [13 f/size/y - 5])

				(
					either connecting [
						compose [
							pen red
							fill-pen none
							box 0x0 (f/size - 1)
							fill-pen red
							polygon
								(to-pair reduce [0 f/size/y - 16])
								(to-pair reduce [16 f/size/y - 0])
								(to-pair reduce [0 f/size/y])
						]
					] [
						[]
					]
				)
			]
		]
	]
	self/repaint		
]

draw-arrow: func [a b color length width parent-face /local d l u n1 n2 p p1 p2] [
	; return compose [pen (color) line (a) (b)]
	d:  b - a
	l:  max 0.01 vector-length d
	u:  reduce [d/1 / l d/2 / l]
	n1: reduce [u/2 (-1 * u/1)] 
	n2: reduce [(-1 * u/2) u/1]
	p:  b - to-pair vector-scale u length
	p1: p + to-pair vector-scale n1 width
	p2: p + to-pair vector-scale n2 width
	if any [ ; @@ revisit-later: /view crashes if polygon outside of visible area
		(not within? p1 5x5 (parent-face/size - 10))
		(not within? p2 5x5 (parent-face/size - 10))
		(not within? b  5x5 (parent-face/size - 10))
		(not within? p  5x5 (parent-face/size - 10))
	] [
		return compose [pen (color) line (a) (b)]
	]
	compose [
		pen (color)
		line (a) (p)
		line (p1) (p2) (b) (p1)
		; fill-pen (color) ; @@ "hard" /view crash prevention ...
		polygon (b) (p1) (p2)
	]
]

check-for-edge-activation: func [p] [
	foreach [a b subject object] edge-coords [
		if 'on-line = point-on-line? a b p [
			either all [(subject = active-edge/1) and (object = active-edge/2)] [
				deactivate-edge
			] [
				active-edge: reduce [subject object]
				deactivate-all-nodes
			]
			workspace/repaint
			show workspace
			exit
		]
	]
]

; -- loading & saving (model dependend) --

load-file: func [file] [
	if none = file [
		file: request-file
		if series? file [
			file: file/1
		]
	]
	if none = file [
		return false
	]
	current-file: file
	space/load current-file 'always
	sync-view
]

save-file: func [file] [
	if none = file [
		file: request-file/save
		if series? file [
			file: file/1
		]
	]
	if none = file [
		return false
	]
	current-file: file
	space/file: current-file
	space/snapshot-behaviour: 'always
	sync-model
	space/snapshot
]

; -- state synchronization (those hooks could - and should - be factored into an interface) --
; @@ currently one-shot hooks (save/load) only, todo: continuous syncing

sync-model: func [/local subjects] [
	; @@ revisit: when full link editing functionality sneaks in
	space/remove-all space/fetch 'query [none (not path? x) none]
	subjects: copy []
	foreach face workspace/pane [
		if face/type = 'node [
			sync-model-node face
			append subjects face/id
		]
	]
	space/remove-all space/fetch 'query [(not found? find subjects x) none none]
	sync-model-edges
]

sync-model-node: func [node] [
	foreach triple compose/deep [
			[(node/id) snip/title (node/text)]
			[(node/id) node/w (node/size/x)]
			[(node/id) node/h (node/size/y)]
			[(node/id) node/x (node/offset/x)]
			[(node/id) node/y (node/offset/y)]
		] [
		space/add triple
	]
]

sync-model-edges: func [/local edge-id] [
	; @@ revisit: when full link editing functionality sneaks in
	foreach edge edges [
		edge-id: generate-id
		space/add reduce  [edge/1/id edge-id edge/2/id]
		space/add compose [(edge-id) edge/type snip/links-to]
	]
]

sync-view: func [/local triples node edge-subjects] [
	deactivate-edge
	deactivate-all-nodes
	clear workspace/pane
	clear edges
	foreach subject space/subjects/for 'node [
		triples: space/fetch 'match reduce [subject]
		node: make node! compose [
			id: (subject)
		]
		node/text: 	get-object triples 'snip/title
		node/offset:	as-pair
					get-object triples 'node/x 
					get-object triples 'node/y
		node/size:	as-pair
					get-object triples 'node/w
					get-object triples 'node/h
		add-complete-node node
	]
	; @@ revisit once link editor exists
	edge-subjects: space/subjects/for 'edge
	foreach edge space/fetch 'query [none (found? find edge-subjects x) none] [
		add-edge
			get-node edge/:subject
			get-node edge/:object
	]
	workspace/repaint
	show workspace
]

; -- state manipulation --

deactivate-edge: does [
	if not empty? active-edge [
		active-edge: copy []
	]
]

deactivate-all-nodes: does [
	foreach face workspace/pane [
		if face/type = 'node [
			if face/active [
				face/active: false
				face/repaint
				show face
			]
		]
	]
	active-node: none
]

activate-node: func [node] [
	if all [(not none? active-node) (active-node/id = node/id)] [
		exit
	]
	deactivate-all-nodes
	deactivate-edge
	remove find workspace/pane node
	append workspace/pane node
	active-node: node
	node/active: true
	node/repaint
	show node
]

add-edge: func [subject object /local edge] [
	edge: reduce [subject object]
	if not found? find/only edges edge [
		append/only edges edge
	]
]

get-node: func [id] [
	foreach face workspace/pane [
		if all [
			face/type = 'node 
			face/id = id
		] [
			return face
		]
	]
	none
]

add-complete-node: func [node] [
	node/repaint
	repend workspace/pane node
	node
]

add-node: func [text position /local node] [
	if none? position [
		position: to-pair reduce [
			random workspace/size/x 
			random workspace/size/y
		]
	]
	add-complete-node make node! compose [
		offset: (position) 
		id: (generate-id) 
		text: (copy text)
	]
]

delete-edge: func [subject object] [
	deactivate-edge
	remove find/only edges reduce [subject object]
]

delete-node: func [node] [
	remove-each edge edges [any [(edge/1 = node) (edge/2 = node)]]
	remove find workspace/pane node
]

view/new/title/options workspace "3flex" [resize]

workspace/feel: make workspace/feel [
	detect: func [f e /local i n] [
		if e/type = 'resize [
			f/delta-to-move: (f/size - f/original-size) / 2
			f/original-size: f/size
			f/hide-arrows
			show f
		]
		if e/type = 'close [
			unview/all
		]
		if e/type = 'key [
			if any [(e/key = #"^~") (e/key = #"^H")] [
				if not empty? active-edge [
					delete-edge active-edge/1 active-edge/2
				]
				if not none? active-node [
					delete-node active-node
				]
				workspace/repaint
				show workspace
			]
			if e/key = #"^-" [
				save/png (i: 0 forever [if not exists? n: to-file rejoin ["3flex-snapshot-" timestamp ".png"] [break/return n] i: i + 1]) to-image workspace
			]
			if e/key = #"^C" [
				print "** Press ESC for the REBOL console."
			]
			if e/key = #"^L" [
				load-file none
			]
			if e/key = #"^S" [
				save-file current-file
			]
		]
		e
	]
	engage: func [f a e] [
		if f/type = 'workspace [
			if a = 'down [
				if 300 > (e/time - f/last-click) [ ; double-click
				 	add-node "New & Exciting" e/offset - (node!/size / 2)
				]
				f/last-click: e/time
				deactivate-all-nodes
				deactivate-edge
				check-for-edge-activation f/data: e/offset
			]
			if a = 'up [
				f/repaint
				show f
			]
			if find [away over] a [
				f/hide-arrows
				foreach face f/pane [
					face/offset: face/offset + (e/offset - f/data)
					show face
				]
				f/data: e/offset
			]
			if a = 'time [
				if not f/delta-to-move = 0x0 [
					if (f/delta-to-move / 2) = 0x0 [
						f/repaint
						show f
					]
					f/delta-to-move: f/delta-to-move / 2
					foreach face f/pane [
						face/offset: face/offset + f/delta-to-move
						show face
					]
				]
			]
		]
	]
]

do-events


