REBOL [
	author: 'chl
	history: [
		2003-01-02 "started development"
		2003-01-08 "adapted to new edge representation concept (edge/type)"
	]
	description: "very simple main-memory based triple-space implementation"
]

space: context [

	export: func [words] [
		foreach word words [
			set in system/words word get word
		]
	]

	subject: 	1
	predicate: 	2
	object: 	3

	export [
		subject predicate object
	]

	file: none
	triples: copy []
	snapshot-behaviour: none ; none or 'always

	load: func [file snapshot-behaviour] [
		self/file: file
		if none? file [
			self/triples: copy []
			return false
		]
		if not exists? file [
			save file copy []
		]
		self/snapshot-behaviour: snapshot-behaviour
		self/triples: system/words/load file
	]

	; layer sub-zero

	to-lit: func [x] [
		if word? x [return to-lit-word x] 
		if path? x [return to-lit-path x]
		x
	]

	replace-deep: func [target search replace /lit] [
		forskip target 1 [
			either (block? target/1) or (paren? target/1) [
				change/only target replace-deep target/1 search replace 
			] [
				if target/1 = search [
					either lit [
						change/only target to-lit replace
					] [
						change/only target replace
					]
				]
			]
		]
		target: head target
	]

	copy-series: func [x] [
		either series? x [
			return copy/deep x
		] [
			return copy reduce [false]
		]
	]

	none?: func [x] [
		(system/words/none? x) or ('none = x)
	]

	unique-triples: func [triples /local new-triples signatures signature] [
		new-triples: copy []
		signatures: copy []
		foreach triple triples [
			if not found? find/only signatures signature: copy/part triple 2 [
				append/only new-triples triple
				append/only signatures signature
			]
		]
		new-triples
	]

	; basic functions

	valid?: func [triple] [
		either any [
			(length? triple) < 3
			not string? triple/:subject
		] [
			false
		] [
			true
		]
	]

	snapshot: does [
		if not none? self/file [
			save self/file self/triples
		]
	]

	snapshot-if-necessary: does [
		if 'always = self/snapshot-behaviour [
			self/snapshot
		]
	]

	add-unchecked: func [triple] [
		insert/only self/triples triple
		triple
	]

	clear: does [
		system/words/clear self/triples
		self/snapshot-if-necessary
		
	]

	add: func [triple] [
		if not valid? triple [
			return false
		]
		if found? find/only self/triples triple [ ; @@ necessary?
			return false
		]
		self/triples: self/unique-triples head insert/only self/triples triple
		self/snapshot-if-necessary
		triple
	]

	remove: func [triple] [
		system/words/remove find/only self/triples triple
		self/snapshot-if-necessary
		triple
	]

	match: func [template f /with-indices /local result] [
		result: copy []
		foreach triple self/triples [
			if all [ ; @@ scrap infix once /view is bug-fixed
			    	any [(self/none? template/:subject) (template/:subject = triple/:subject)]
			    	any [(self/none? template/:predicate) (template/:predicate = triple/:predicate)]
			    	any [(self/none? template/:object) (template/:object = triple/:object)] 
			] [
				append/only result triple
			]
		]
		foreach r result [
			f r
		]
		none
	]

	; space/query [none [(x = "Vanilla") or (x = "chris")] none none]

	query: func [template f /local result] [
		result: copy []
		foreach triple triples [
			if all [
			    	any [(self/none? template/:subject) (do self/replace-deep/lit self/copy-series template/:subject 'x triple/:subject)]
				any [(self/none? template/:predicate) (do self/replace-deep/lit self/copy-series template/:predicate 'x triple/:predicate)]
				any [(self/none? template/:object) (do self/replace-deep/lit self/copy-series template/:object 'x triple/:object)]
			] [
				append/only result triple
			]
		]
		foreach r result [
			f r
		]
		none
	]

	; shortcuts & sugar

	exists-subject?: func [subject] [
		found? find self/subjects subject
	]

	subjects: func [/local subjects /for category] [
		subjects: copy []
		foreach triple self/triples [
			if any [
				not for
				all [
					path? triple/:predicate
					category = first triple/:predicate 
				]
			] [
				append subjects triple/:subject
			]
		]
		unique subjects
	]

	build: func [method template f /local triples] [
		triples: copy []
		self/:method template func [triple] [
			append/only triples f triple
		]
		triples
	]

	fetch: func [method template] [
		self/build method template func [x] [x]
	]

	probe: func [method template] [
		self/:method template func [x] [system/words/probe x]
	]

	for-all: func [triples method /local fomer-behaviour] [
		former-behaviour: self/snapshot-behaviour
		self/snapshot-behaviour: none
		foreach triple triples [
			self/:method triple
		]
		self/snapshot-behaviour: former-behaviour
		self/snapshot-if-necessary
		self/triples
	]

	add-all: func [triples /local fomer-behaviour] [
		append self/triples triples
		self/triples: unique-triples self/triples
		self/snapshot-if-necessary
	]

	remove-all: func [triples /local fomer-behaviour] [
		self/for-all triples 'remove
	]

	get-object: func [subject predicate] [
		foreach triple self/triples [
			if all [
				triple/:subject = subject
				triple/:predicate = predicate
			] [
				return triple/:object
			]
		]
	]

]

if found? find to-string system/options/script %3space-in-memory-lib.r [
	halt
]

