REBOL [
	author: 'chl
	date: 2003-01-02
]

random/seed now/time/precise

golden-section: (1 + square-root 5) / 2

golden: func [x] [
	x - (x / golden-section)
]

luma: func [color factor /local new-color i] [
	new-color: copy [] 
	for i 1 3 1 [
		append new-color min 255 to-integer color/:i * factor
	]
	make tuple! new-color
]

vector-length: func [v] [
	square-root (v/1 ** 2) + (v/2 ** 2)
]

vector-scale: func [v factor] [
	to type? v reduce [
		to-integer factor * v/1
		to-integer factor * v/2
	]
]

min~: func [x /local m] [m: x/1 foreach e x [m: min m e] m]

initial-workspace-size: make pair! reduce [
	to-integer (0.8 * system/view/screen-face/size/x)
	to-integer (0.8 * system/view/screen-face/size/x) / golden-section
]

generate-id: func [/local r] [
	r: copy "-"
	loop 4 [repend r [random 210000000 "-"]]
	r
]

pad-2: func [x] [x: to-string x either (length? x) = 1 [join "0" x] [x]]

timestamp: func [/local d t] [
	d: now/date
	t: now/time
	rejoin [
		d/year pad-2 d/month pad-2 d/day pad-2 t/hour pad-2 t/minute pad-2 t/second
	]
]

point-on-line?: func [a b c /local qx qy px py tx ty] [
	"by Alan Paeth, from Graphics Gems, Academic Press, 1990"
	qx: a/x qy: a/y px: b/x py: b/y tx: c/x ty: c/y
	if ((abs (((qy - py) * (tx - px)) - ((ty - py) * (qx - px)))) >
 ((abs (qx - px)) + (abs (qy - py)))) [return false]
	if (((qx < px) and (px < tx)) or ((qy < py) and (py < ty))) [return 'on-ray-a]
	if (((tx < px) and (px < qx)) or ((ty < py) and (py < qy))) [return 'on-ray-a]
	if (((px < qx) and (qx < tx)) or ((py < qy) and (qy < ty))) [return 'on-ray-b]
	if (((tx < qx) and (qx < px)) or ((ty < qy) and (qy < py))) [return 'on-ray-b]
	return 'on-line
]

get-object: func [triples predicate-q] [ ; @@ KILL
	foreach triple triples [
		if triple/:predicate = predicate-q [
			return triple/:object
		]
	]
	none
]
