REBOL [
	author: 'chl
	history: [
		2002-01-04	"started development"
		2002-01-05	"look and feel"
		2002-01-06	"finished first working prototype"
		2002-01-08	"adapted to new 3space"
	]
]

either found? find to-string system/options/script %node-editor [
	do %misc-lib.r
	do %3space-in-memory-lib.r
	space/load %../files/test.3flex none
	default-font-small: make face/font [
		name: "Verdana"
		size: 11
	]
	prefs: context [
		drawer-header-height: 20
		drawer-color: 240.240.240
	]
	libs: %./
] [
	libs: %lib/
]

node-editor: context [

	; -- interface --
	
	init: func [space] [
		self/space: space
	]
	
	switch-subject: func [subject] [
		node-editor-shell/init self/space subject
		node-editor-shell/repaint
		either not found? find system/view/screen-face/pane node-editor-shell [
			either none? node-editor-shell/last-offset [
				center-face node-editor-shell
			] [
				node-editor-shell/offset: node-editor-shell/last-offset
			]
		] [
			unview/only node-editor-shell
		]
		view/new/title/options node-editor-shell "Edit Node" [resize]
		node-editor-shell/reset-feel
		show node-editor-shell
	]

	add-change-listener: func [listener] [
		append change-listeners :listener
	]

	global-notify-change: func [subject] [
		foreach listener change-listeners [
			listener subject
		]
	]

	; -- implementation --

	change-listeners: copy []

	space: none

	last-offset: none

	arrow-images: reduce [load %../assets/arrow.png load %../assets/arrow-rotated.png]
	
	make-arrow: does [
		make face [
			image: first arrow-images
			size: self/image/size
			edge: make self/edge [size: 0x0]
			color: prefs/drawer-color
		]
	]
	
	drawer-header: make face [
		size: as-pair 400 prefs/drawer-header-height
		offset: 0x0
		edge: make self/edge [size: 0x0]
		color: prefs/drawer-color
		text: "ZORK!"
		font: make font [
			name: "Verdana"
			style: 'bold
			size: 11
			color: black
			shadow: none
			align: 'left
		]
		para: make para [
			origin: 20x3
		]
		arrow: none
		drawer: none
		init: does [
			self/arrow: make-arrow
			self/pane: reduce [self/arrow]
		]
		repaint: does [
			self/effect: compose/deep [
				draw [
					pen white
					line 1x1 (as-pair 1 (self/size/y - 1))
					line 1x1 (as-pair (self/size/x - 1) 1)
					pen (luma prefs/drawer-color 0.75)
					line (as-pair (self/size/x - 1) 1) (self/size - 1)
					line (as-pair 1 (self/size/y - 1)) (self/size - 1)
					pen (luma prefs/drawer-color 0.75)
					line 
						(as-pair max 100 0.5 * golden self/size/x 2) 
						(as-pair max 100 0.5 * golden self/size/x self/size/y - 2)
				]
			]
			if not none? self/drawer [
				self/arrow/image: do either self/drawer/open ['second] ['first] arrow-images
				self/arrow/offset: either self/drawer/open [7x7] [8x6]
			]
		]
		feel: make feel [
			engage: func [f a e] [
				switch a [
					down [
						f/drawer/open: not f/drawer/open
						f/drawer/shell/repaint
						show f/drawer/shell
					]
				]
			]
		]
	]
	
	drawer: make face [
		edge: make self/edge [size: 0x0]
		pane: copy []
		open: true
		shell: none
		category: none
		header: none
		category-editor: none
		init: func [shell category category-editor /local category-spec] [
			self/shell: shell
			self/category: category
			self/open: get in (self/shell/load-category-spec self/category) 'auto-open
			clear pane
			repend pane [
				self/header: make drawer-header compose [
					text: (to-string category)
					drawer: (self)
				]
				self/category-editor: category-editor
			]
			self/category-editor/offset/y: prefs/drawer-header-height
			self/header/init
			
		]
		repaint: does [
			self/set-x self/size/x
			self/size/y: self/header/size/y + (either self/open [self/category-editor/size/x] [0])
			self/header/repaint
		]
		set-x: func [x] [
			self/category-editor/set-x self/size/x: self/header/size/x: x
		]
	]
	
	node-editor-shell: make face [
		size: 360x1
		edge: make self/edge [size: 0x0]
		pane: copy []
		space: none
		subject: none
		triples: copy [] ; @@ do we really need that? heck it's convenient ;-)
		last-offset: none
		reset-feel: does [
			self/feel: make self/feel [
				detect: func [f e] [
					if e/type = 'resize [
						f/repaint
						show f
					]
					if e/type = 'close [
						f/last-offset: f/offset
						f/notify-change f/subject
					]
				e
				]
			]
		]
		load-category-spec: func [category] [
			do load rejoin [libs %categories/ category ".category"]
		]
		reset-x: does [
			foreach face self/pane [
				face/set-x self/size/x
			]
		]
		recalc-y: func [/local y editor] [
			self/size/y: 1
			foreach face self/pane [
				face/offset/y: self/size/y
				self/size/y: self/size/y + prefs/drawer-header-height + (either face/open [face/category-editor/size/y] [0])
			]
		]
		repaint: does [
			self/reset-x
			self/recalc-y
			foreach face self/pane [
				face/repaint
			]
		]
		init: func [space subject /local categories predicates d file e] [
			if not empty? pane [
				clear self/pane
			]
			self/space: space
			self/subject: subject
			categories: copy []
			predicates: copy []
			foreach triple self/triples: space/fetch 'match reduce [subject] [
				if path? triple/:predicate [
					append categories first triple/:predicate
					append predicates triple/:predicate
				]
			]
			categories: unique categories
			predicates: unique predicates
			foreach category categories [
				d: make drawer []
				either exists? file: rejoin [libs %category-editors/ category ".category-editor"] [
					e: get in do load file 'editor
				] [
					e: get in do load join libs %category-editors/default.category-editor 'editor
				]
				d/init self category e
				e/init d
				append self/pane d
			]
		]
		notify-change: func [subject] [
			global-notify-change subject
		]
	]
	
]

if found? find to-string system/options/script %node-editor [
	node-editor/init space
	node-editor/switch-subject first space/subjects/for 'node
	do-events
]

